package org.xbib.logging.handlers;

import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.util.concurrent.locks.ReentrantLock;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import org.xbib.logging.formatters.ThreadLoggingFormatter;

/**
 * Replacement for java.util.logging.StreamHandler that is wired to the ThreadLoggingFormatter
 * and does not use any fall-back formatter
 */
public class ThreadLoggingStreamHandler extends Handler {

    private static final ThreadLoggingFormatter formatter = new ThreadLoggingFormatter();

    private final ReentrantLock lock;

    private OutputStream output;

    private boolean doneHeader;

    private volatile Writer writer;

    public ThreadLoggingStreamHandler() {
        super();
        this.lock = this.initLocking();
        setLevel(Level.ALL);
        setFilter(null);
        setFormatter(formatter);
        try {
            setEncoding("UTF-8");
        } catch (UnsupportedEncodingException e) {
            // can't happen
        }
    }

    protected void setOutputStream(OutputStream out) {
        if (this.tryUseLock()) {
            try {
                this.setOutputStream0(out);
            } finally {
                this.unlock();
            }
        } else {
            synchronized (this) {
                this.setOutputStream0(out);
            }
        }
    }

    private void setOutputStream0(OutputStream out) {
        if (out == null) {
            throw new NullPointerException();
        } else {
            this.flushAndClose();
            this.output = out;
            this.doneHeader = false;
            String encoding = this.getEncoding();
            if (encoding == null) {
                this.writer = new OutputStreamWriter(this.output);
            } else {
                try {
                    this.writer = new OutputStreamWriter(this.output, encoding);
                } catch (UnsupportedEncodingException e) {
                    throw new Error("Unexpected exception " + e);
                }
            }
        }
    }

    @Override
    public void setEncoding(String encoding) throws UnsupportedEncodingException {
        if (this.tryUseLock()) {
            try {
                this.setEncoding0(encoding);
            } finally {
                this.unlock();
            }
        } else {
            synchronized (this) {
                this.setEncoding0(encoding);
            }
        }
    }

    private void setEncoding0(String encoding) throws UnsupportedEncodingException {
        super.setEncoding(encoding);
        if (this.output != null) {
            this.flush();
            if (encoding == null) {
                this.writer = new OutputStreamWriter(this.output);
            } else {
                this.writer = new OutputStreamWriter(this.output, encoding);
            }
        }
    }

    @Override
    public void publish(LogRecord record) {
        if (this.tryUseLock()) {
            try {
                this.publish0(record);
            } finally {
                this.unlock();
            }
        } else {
            synchronized (this) {
                this.publish0(record);
            }
        }
    }

    private void publish0(LogRecord record) {
        if (this.isLoggable(record)) {
            String msg;
            try {
                msg = this.getFormatter().format(record);
            } catch (Exception e) {
                this.reportError(null, e, 5);
                return;
            }
            try {
                if (!this.doneHeader) {
                    this.writer.write(this.getFormatter().getHead(this));
                    this.doneHeader = true;
                }
                this.writer.write(msg);
            } catch (Exception e) {
                this.reportError(null, e, 1);
            }
        }
    }

    @Override
    public boolean isLoggable(LogRecord record) {
        return this.writer != null && record != null && super.isLoggable(record);
    }

    @Override
    public void flush() {
        if (this.tryUseLock()) {
            try {
                this.flush0();
            } finally {
                this.unlock();
            }
        } else {
            synchronized (this) {
                this.flush0();
            }
        }
    }

    private void flush0() {
        if (this.writer != null) {
            try {
                this.writer.flush();
            } catch (Exception e) {
                this.reportError(null, e, 2);
            }
        }
    }

    private void flushAndClose() {
        if (this.writer != null) {
            try {
                if (!this.doneHeader) {
                    this.writer.write(this.getFormatter().getHead(this));
                    this.doneHeader = true;
                }
                this.writer.write(this.getFormatter().getTail(this));
                this.writer.flush();
                this.writer.close();
            } catch (Exception e) {
                this.reportError(null, e, 3);
            }
            this.writer = null;
            this.output = null;
        }
    }

    @Override
    public void close() {
        if (this.tryUseLock()) {
            try {
                this.flushAndClose();
            } finally {
                this.unlock();
            }
        } else {
            synchronized (this) {
                this.flushAndClose();
            }
        }
    }

    private ReentrantLock initLocking() {
        Class<?> clazz = this.getClass();
        ClassLoader loader = clazz.getClassLoader();
        return loader != null && loader != ClassLoader.getPlatformClassLoader() ? null : new ReentrantLock();
    }

    private boolean tryUseLock() {
        if (this.lock == null) {
            return false;
        } else {
            this.lock.lock();
            return true;
        }
    }

    private void unlock() {
        this.lock.unlock();
    }
}
